#include "aliases/aliases.h"
#include "nodes.h"
#include "ops.h"
#include "ui.h"
#include "window.h"
#include <SDL2/SDL.h>
#include <SDL2/SDL_events.h>
#include <SDL2/SDL_render.h>
#include <SDL2/SDL_video.h>
#include <stdbool.h>
#include <stdlib.h>

#define MAX_WINDOWS 2

#define WINDOW_WIDTH 1280
#define WINDOW_HEIGHT 720

typedef struct compositor compositor_t;
struct compositor {
  window_t windows[MAX_WINDOWS];
  u32 active_window;
  SDL_Event event;
  bool running;
  u64 mouse_x;
  u64 mouse_y;
  u64 last_clicked_mouse_x;
  u64 last_clicked_mouse_y;
  i64 node_hovered;
  u64 count;
  node_t *nodes;
  bool move_node;
  ui_ctx_t ctx;
};

void add_node(compositor_t *comp, node_type_t type, node_data_t data, i32 x,
              i32 y);

i32 run_main_loop(void) {
  if (SDL_Init(SDL_INIT_EVERYTHING) != 0) {
    return EXIT_FAILURE;
  }

  compositor_t comp = {0};

  init_ui_ctx(&(comp.ctx));

  comp.nodes = (node_t *)malloc(sizeof(node_t) * MAX_NODES);

  window_t *main_window = &(comp.windows[0]);
  window_t *toolbox = &(comp.windows[1]);

  if (!init_window(main_window, "Compositor", WINDOW_WIDTH, WINDOW_HEIGHT, -1,
                   -1)) {
    SDL_Quit();

    return EXIT_FAILURE;
  }

  u32 toolbox_window_width = WINDOW_WIDTH / 7;
  init_window(toolbox, "Toolbox", toolbox_window_width, WINDOW_HEIGHT,
              main_window->x - toolbox_window_width, -1);

  comp.running = true;

  SDL_EventState(SDL_DROPFILE, SDL_ENABLE);

  colour_t bg_colour = {.abgr = 0xffffffff};

  while (comp.running) {
    while (SDL_PollEvent(&(comp.event))) {
      handle_ui_events(&(comp.windows[comp.active_window - 1]), &(comp.ctx),
                       &(comp.event));

      switch (comp.event.type) {
      case SDL_QUIT:
        comp.running = false;
        break;
      case SDL_WINDOWEVENT:
        switch (comp.event.window.event) {
        case SDL_WINDOWEVENT_CLOSE:
          comp.running = false;
          break;
        case SDL_WINDOWEVENT_ENTER: {
          u32 id = comp.event.window.windowID;
          window_t *wnd = NULL;

          for (u64 i = 0; i < MAX_WINDOWS; ++i) {
            window_t *window = &(comp.windows[i]);

            if (id == window->id) {
              comp.active_window = id;
              wnd = window;
              break;
            }
          }

          if (!wnd) {
            break;
          }

          SDL_RaiseWindow(wnd->window);

          break;
        }
        }

        break;
      }
    }

    for (u64 i = 0; i < MAX_WINDOWS; ++i) {
      clear_window(&(comp.windows[i]), bg_colour);
    }

    for (u64 i = 0; i < MAX_WINDOWS; ++i) {
      swap_buffers(&(comp.windows[i]));
    }

    reset_ui_ctx(&(comp.ctx));
  }

#if 0
  i32 button_x = (toolbox->width - BUTTON_WIDTH) / 2;
  for (u64 i = 0; i < COUNT_COMP_OPS; ++i) {
    comp.buttons[i] = (button_t){.rect = (rect_t){
                                     .topleft.x = button_x,
                                     .topleft.y = i * (BUTTON_HEIGHT + 20) + 30,
                                     .w = BUTTON_WIDTH,
                                     .h = BUTTON_HEIGHT,
                                 }};
  }

  while (comp.running) {
    while (SDL_PollEvent(&(comp.event))) {
      switch (comp.event.type) {
      case SDL_QUIT:
        comp.running = false;
        break;
      case SDL_WINDOWEVENT:
        switch (comp.event.window.event) {
        case SDL_WINDOWEVENT_CLOSE:
          comp.running = false;
          break;
        case SDL_WINDOWEVENT_ENTER: {
          u32 id = comp.event.window.windowID;
          window_t *wnd = NULL;

          for (u64 i = 0; i < MAX_WINDOWS; ++i) {
            window_t *window = &(comp.windows[i]);

            if (id == window->id) {
              wnd = window;
              break;
            }
          }

          if (!wnd) {
            break;
          }

          SDL_RaiseWindow(wnd->window);

          break;
        }
        }

        break;
      case SDL_MOUSEBUTTONDOWN:
        if (comp.event.button.windowID == main_window->id) {
          if (comp.node_hovered != -1) {
            comp.move_node = true;
          }

          comp.last_clicked_mouse_x = comp.event.button.x;
          comp.last_clicked_mouse_y = comp.event.button.y;

          break;
        } else if (comp.event.button.windowID == toolbox->id) {
          if (comp.button_hovered != -1) {
            comp.button_clicked = comp.button_hovered;
          }

          break;
        }
      case SDL_MOUSEBUTTONUP:
        comp.move_node = false;

        if (comp.event.button.windowID == toolbox->id) {
          if (comp.button_hovered >= 0 &&
              comp.button_hovered == comp.button_clicked) {
            add_node(&comp, NODE_TYPE_OP,
                     (node_data_t){.func = ops[comp.button_hovered]},
                     comp.last_clicked_mouse_x, comp.last_clicked_mouse_y);
          }

          comp.button_clicked = -1;
        }

        break;
      case SDL_MOUSEMOTION:
        comp.button_hovered = -1;

        if (comp.event.motion.windowID == main_window->id) {
          comp.mouse_x = comp.event.motion.x;
          comp.mouse_y = comp.event.motion.y;

          if (comp.move_node) {
            i32 dx = comp.event.motion.xrel;
            i32 dy = comp.event.motion.yrel;

            node_t *node = &(comp.nodes[comp.node_hovered]);

            node->rect.topleft.x += dx;
            node->rect.topleft.y += dy;
          } else {
            comp.node_hovered = -1;

            for (u64 i = comp.count - 1; i >= 0; --i) {
              rect_t *rect = &(comp.nodes[i].rect);

              if (aabb(rect, comp.mouse_x, comp.mouse_y)) {
                comp.node_hovered = i;
                break;
              }
            }
          }

          break;
        } else if (comp.event.motion.windowID == toolbox->id) {
          for (u64 i = COUNT_COMP_OPS - 1; i >= 0; --i) {
            rect_t *rect = &(comp.buttons[i].rect);

            if (aabb(rect, comp.event.motion.x, comp.event.motion.y)) {
              comp.button_hovered = i;
              break;
            }
          }

          break;
        }
      case SDL_DROPFILE:
        if (comp.event.drop.windowID == main_window->id) {
          node_data_t data = (node_data_t){.path = comp.event.drop.file};

          add_node(&comp, NODE_TYPE_IO, data, comp.mouse_x, comp.mouse_y);

          break;
        }
      }
    }

    for (u64 i = 0; i < MAX_WINDOWS; ++i) {
      clear_window(&(comp.windows[i]), bg_colour);
    }

    for (u64 i = 0; i < comp.count; ++i) {
      node_t *node = &(comp.nodes[i]);
      draw_node(main_window, node);
    }

    for (u64 i = 0; i < COUNT_COMP_OPS; ++i) {
      button_t *button = &(comp.buttons[i]);
      draw_button(toolbox, button);
    }

    for (u64 i = 0; i < MAX_WINDOWS; ++i) {
      swap_buffers(&(comp.windows[i]));
    }
  }
#endif

  for (u64 i = 0; i < MAX_WINDOWS; ++i) {
    cleanup_window(&(comp.windows[i]));
  }

  SDL_Quit();

  return EXIT_SUCCESS;
}

#if 0
void add_node(compositor_t *comp, node_type_t type, node_data_t data, i32 x,
              i32 y) {
  if (comp->count + 1 >= MAX_NODES) {
    return;
  }

  comp->nodes[(comp->count)++] = (node_t){
      .rect =
          (rect_t){
              .topleft.x = x,
              .topleft.y = y,
              .w = NODE_WIDTH,
              .h = NODE_HEIGHT,
          },
      .type = type,
      .data.path = data.path,
  };
}
#endif
